/*
 * I2Cdriver.xc
 *
 *  Created on: 29 Aug 2011
 *      Author: Martin
 */

#include <platform.h>
#include <xclib.h>
#include "I2Cdriver.h"

//----------------------------------------------
// assign ports specified in configuration file
//----------------------------------------------
I2C_PORTSPEC_SCL
I2C_PORTSPEC_SDA
I2C_PORTSPEC_CLK


//----------------------------------
// global variable (internal scope)
//----------------------------------
int count;											// port counter


//----------------------------------
// prototypes for private functions
//----------------------------------
static void	I2Cstart				( void );
static void	I2Cstop					( void );
static void I2CrepeatStart			( void );
static char	I2Caddress				( unsigned char address, unsigned char rw );
static char	I2CsendByte				( unsigned char byte );
static unsigned char I2CreadByte	( unsigned char ackResponse );


//------------------
// public functions
//------------------
void I2Cinit( void ) {

	configure_clock_rate(I2C_CLK, I2C_TIMING_CLK_NUMERATOR, I2C_TIMING_CLK_DENOMINATOR);

	set_port_pull_up(I2C_PORT_SCL);
	configure_out_port(I2C_PORT_SCL, I2C_CLK, 1);

	set_port_pull_up(I2C_PORT_SDA);
	configure_out_port(I2C_PORT_SDA, I2C_CLK, 1);

	start_clock(I2C_CLK);
}


void I2Csync( void ) {
	sync(I2C_PORT_SDA);
}


char I2CwriteReg8( unsigned char busAddr, unsigned char regAddr, unsigned char data ) {
	// write 8 bit register
	// Returns: 0 if a peripheral acknowleged, otherwise -1

	I2Cstart();
	if ( I2Caddress(busAddr, I2C_WRITE) )
		return -1;
	if ( I2CsendByte(regAddr) )
		return -1;
	if ( I2CsendByte(data) )
		return -1;
	I2Cstop();

	return 0;
}


char I2CreadReg8( unsigned char busAddr, unsigned char regAddr, unsigned char &data ) {
	// read 8 bit register
	// Returns: 0 if a peripheral acknowledged, otherwise -1

	// address peripheral and write register address
	I2Cstart();
	if ( I2Caddress(busAddr, I2C_WRITE) )
		return -1;
	if ( I2CsendByte(regAddr) )
		return -1;

	// re-address peripheral and read register
	I2CrepeatStart();
	if ( I2Caddress(busAddr, I2C_READ) )
		return -1;
	data = I2CreadByte(I2C_NACK);			// send NACK repsonse to indicate no more data required
	I2Cstop();

	return 0;
}


char I2CwriteReg16( unsigned char busAddr, unsigned char regAddr, unsigned data ) {
	// write 16 bit register
	// Returns: 0 if a peripheral acknowleged, otherwise -1

	I2Cstart();
	if ( I2Caddress(busAddr, I2C_WRITE) )
		return -1;
	if ( I2CsendByte(regAddr) )
		return -1;
	if ( I2CsendByte(data >> 8) )
		return -1;
	if ( I2CsendByte(data & 0xff) )
		return -1;
	I2Cstop();

	return 0;
}


char I2CreadReg16( unsigned char busAddr, unsigned char regAddr, unsigned &data ) {
	// read 16 bit register
	// Returns: 0 if a peripheral acknowleged, otherwise -1

	I2Cstart();
	if ( I2Caddress(busAddr, I2C_WRITE) )	// address peripheral and initiate write transaction
		return -1;
	if ( I2CsendByte(regAddr) )				// write address register
		return -1;
	I2CrepeatStart();						// re-address peripheral and initiate read transaction
	if ( I2Caddress(busAddr, I2C_READ) )
		return -1;
	data = I2CreadByte( I2C_ACK ) << 8;		// read high byte and send ACK to indicate more data required
	data |= I2CreadByte( I2C_NACK );		// read low byte and send NACK to indicate no more data required

	I2Cstop();

	return 0;
}



//-------------------
// private functions
//-------------------
static void I2Cstart( void ) {
	// send master start: SDA falling edge during SCL high period
	extern int count;

	I2C_PORT_SCL <: 1@count;			// read port clock
	count+=2;
	I2C_PORT_SDA@count <: 0;			// create SDA falling edge
	count+=1;
	I2C_PORT_SCL@count <: 0;			// take SCL low
}


static void I2CrepeatStart( void ) {
	// send master repeat start
	extern int count;

	count+=1;
	I2C_PORT_SDA@count <: 1;			// middle of clock low period: set SDA high
	count+=1;
	I2C_PORT_SCL@count <: 1;			// end of clock low period: take SCL high
	count+=1;
	I2C_PORT_SDA@count <: 0;			// middle of clock high period: create SDA falling edge
	count+=1;
	I2C_PORT_SCL@count <: 0;			// end of clock high period: take SCL low
}


static void I2Cstop( void ) {
	// send master stop: SDA rising edge during SCL high period
	extern int count;

	count+=1;
	I2C_PORT_SDA@count <: 0;			// halfway through clock low period: take SDA low
	count+=1;
	I2C_PORT_SCL@count <: 1;			// end of clock low period: take SCL high
	count+=1;
	I2C_PORT_SDA@count <: 1;			// create SDA rising edge
}


static char I2Caddress( unsigned char address, unsigned char rw ) {
	// send start then address device
	// entry state: SCL=1, SDA=1
	// exit state: SCL=0, SDA=1
	// Returns: I2C_STATUS_SUCCESS if the peripheral acknowleged, otherwise I2C_STATUS_FAILURE

	// set up 8 bit data word: peripheral address (big endian) followed by r/w flag
	address <<= 1;						// move address into upper seven bits
	if( rw != I2C_WRITE ) {				// append r/w flag
		address |= 0x01;
	}
	return I2CsendByte(address);
}


static char I2CsendByte(unsigned char byte) {
	// send 8 bit data word
	// Return: 0 if a peripheral acknowledged the byte, otherwise -1
	extern int count;
	unsigned data, ack;

	data = byterev(bitrev(byte));		// convert to big-endian format

	for ( int i=0; i<8; i++ ) {
		count+=1;
		I2C_PORT_SDA@count <: >> data;	// halfway through clock low period: set SDA to next data bit
		count+=1;
		I2C_PORT_SCL@count <: 1;		// end of clock low period: take SCL high
		count+=1;
		I2C_PORT_SCL@count <: 0;		// end of clock high period: take SCL low
	}

	// check response (peripheral responds by pulling SDA low during low period before 9th clock pulse)
	count+=1;
	I2C_PORT_SDA@count <: 1;			// halfway through clock low period: stop pulling SDA low
	count+=1;
	I2C_PORT_SCL@count <: 1;			// end of clock low period: take SCL high
	sync(I2C_PORT_SDA);
	ack = peek(I2C_PORT_SDA);			// sample data while clock is high
	count+=1;
	I2C_PORT_SCL@count <: 0;			// end of clock high period: take SCL low

#ifdef I2C_FAKE_ACK
	return 0;
#endif

	// return peripheral ACK status
	if ( ack == 0 )
		return 0;
	else
		return -1;
}


static unsigned char I2CreadByte( unsigned char ackResponse ) {
	// read 8 bit data word and respond with ACK (more data required) or NACK (no more data required)
	// Returns: the data that was read
	extern int count;
	unsigned data = 0;

	for( int i=0; i<8; i++ ) {
		count+=1;
		I2C_PORT_SDA@count <: 1;		// halfway through clock low period: stop pulling SDA low
		count+=1;
		I2C_PORT_SCL@count <: 1;		// end of clock low period: take SCL high
		sync(I2C_PORT_SDA);
		data <<= 1;
		data |= peek(I2C_PORT_SDA);		// read next bit of data
		count+=1;
		I2C_PORT_SCL@count <: 0;		// end of clock high period: take SCL low
	}

	// send ACK or NACK to indicate whether more data is required
	count+=1;
	I2C_PORT_SDA@count <: ackResponse;	// middle of clock low period: set up required ACK response
	count+=1;
	I2C_PORT_SCL@count <: 1;			// end of clock low period: take SCL high
	count+=1;
	I2C_PORT_SCL@count <: 0;			// end of clock high period: take SCL low

	return data;						// NB: no need to reverse bits as MSB was read first
}


