/*
 * LCDdriver.xc
 *
 *  Created on: 28 Jun 2011
 *      Author: Martin
 *
 *     Purpose: functions to drive LCD display - configuration in LCDdriver.h
 */

#include <platform.h>
#include "LCDdriver.h"


// assign ports specified in configuration file
LCD_PORTSPEC_RS
LCD_PORTSPEC_RW
LCD_PORTSPEC_E
LCD_PORTSPEC_DB


char LCDbusyWait ( void ) {
	// wait for busy flag to be clear, timeout after Tbusymax (x 10ns, 0=disabled)
	// Returns: address counter or LCD_RESULT_TIMEOUT

	timer 		tmrA, tmrB;
	unsigned 	t, tmax;
	char 		status;

	tmrB :> tmax;													// set overall timeout
	tmax += LCD_TIMING_Tbusytimeout;

	// set up read status operation
	LCD_PORT_DB :> void;											// configure data port for input
	LCD_PORT_RS <: LCD_TYPE_INSTR;									// register select: instruction
	LCD_PORT_RW <: LCD_TYPE_READ;									// data direction: read

	// read status, loop while busy; abort on timeout
	do {
		LCD_PORT_E <: 1;											// enable: pulse (rising edge)

		tmrA :> t;													// wait: enable pulse width or timeout
		select {
			case tmrA when timerafter(t + LCD_TIMING_Tpw) :> void :
				break;
			case LCD_TIMING_Tbusytimeout => tmrB when timerafter(tmax) :> void :
				return LCD_RESULT_TIMEOUT;
		}

		LCD_PORT_DB :> status;										// read AC and Busy flag

		LCD_PORT_E  <: 0;											// enable: clock (falling edge)

		tmrA :> t;													// wait: remaining enable cycle time or timeout
		select {
			case tmrA when timerafter(t + LCD_TIMING_Tc - LCD_TIMING_Tpw) :> void :
				break;
			case LCD_TIMING_Tbusytimeout => tmrB when timerafter(tmax) :> void :
				return LCD_RESULT_TIMEOUT;
		}

	} while ( status & LCD_MASK_BUSY );								// loop while busy

	return status;
}


char LCDwrite( const unsigned char data ) {
	// write character to LCD at cursor position, first checking busy flag with timeout (x 10ns, 0=disabled)
	// Returns: address counter or LCD_RESULT_TIMEOUT

	// TODO: work out what happens when user writes beyond the end of the display

	unsigned char status, AC;

	status = LCDbusyWait();											// wait for busy flag to be clear, read address counter
	if ( status == LCD_RESULT_TIMEOUT )
		return LCD_RESULT_TIMEOUT;
	else
		AC = status;

	switch ( data ) {												// convert recognised control codes into instructions

		// backspace
		case '\b' :
			if ( AC != 0x00 ) {
				AC--;												// decrement AC (but not below zero) TODO: scroll display down if this happens?
				status = LCDcommand( 0x80 | AC);
			}
			break;

		// formfeed
		case '\f' :
			status = LCDcommand( 0x01 );							// clear screen
			break;

		// newline
		case '\n' :
			if ( AC >= 0x40 )										// cursor already on second line: scroll
				if ( LCDverticalScroll() == LCD_RESULT_TIMEOUT )
					return LCD_RESULT_TIMEOUT;
			AC = 0x40;												// move to left hand side of second line
			status = LCDcommand( 0x80 | AC );
			break;

		// carriage return
		case '\r' :
			if ( AC < 0x40 )									// cursor on first line
				AC = 0x00;										// move to left hand side of first line
			else
				AC = 0x40;										// move to left hand side of second line
			status = LCDcommand( 0x80 | AC );
			break;

		// horizontal tab
		case '\t' :
			do
				AC ++;
			while( AC % LCDtabWidth );							// advance to next tab stop
			if ( AC > 0x7f )									// don't go off end of display	TODO: scroll if this happens
				AC = 0x7f;
			status = LCDcommand( 0x80 | AC );
			break;

		// vertical tab
		case '\v' :
			if ( AC < 0x40 )									// cursor on first line
				AC += 0x40;
			else												// cursor on second line: scroll
				if ( LCDverticalScroll() == LCD_RESULT_TIMEOUT )
					return LCD_RESULT_TIMEOUT;
			status = LCDcommand( 0x80 | AC );
			break;

		// normal character
		default :
			LCDoutRaw(LCD_TYPE_DATA, data);						// TODO: scroll if user outputs past last char of display
			break;
	}

	return status;
}


char LCDmsg( const unsigned char msg[] ) {
	// write message to LCD at current cursor position, checking busy flag
	// timeout is currently not used
	// Returns: address counter or LCD_RESULT_TIMEOUT
	char status = 0;

	for ( int i=0; msg[i] != '\0' && status != LCD_RESULT_TIMEOUT; i++ )
		status = LCDwrite( msg[i] );

	return status;
}


char LCDcommand( const unsigned char opcode ) {
	// send command to LCD, first checking busy flag with timeout (x 10ns, 0=disabled)
	// Returns: address counter or LCD_RESULT_TIMEOUT

	char status;

	status = LCDbusyWait();
	if ( status != LCD_RESULT_TIMEOUT )
		LCDoutRaw(LCD_TYPE_INSTR, opcode);

	return status;
}


void LCDoutRaw( const unsigned char rs, const unsigned char byte ) {
	// send character or data to LCD without checking busy flag
	// rs=0: instruction, rs=1: data
	timer tmr;
	unsigned t;

	LCD_PORT_RS <: rs;												// register select: 1=data, 0=instruction
	LCD_PORT_RW <: LCD_TYPE_WRITE;									// data direction: write

	LCD_PORT_E  <: 1;												// enable: pulse (rising edge)
	LCD_PORT_DB <: byte;											// output data on port
	tmr :> t;														// wait: enable pulse width
	tmr when timerafter(t + LCD_TIMING_Tpw) :> void;

	LCD_PORT_E  <: 0;												// enable: clock (falling edge)
	tmr :> t;														// wait: remaining enable cycle time
	tmr when timerafter(t + LCD_TIMING_Tc - LCD_TIMING_Tpw) :> void;
}


char LCDverticalScroll( void ) {
	// scroll display up by one line

	timer tmr;
	unsigned t;
	unsigned char row[16];

	if ( LCDcommand(0x80 | 0x40) == LCD_RESULT_TIMEOUT )			// set AC to first character of second line
		return LCD_RESULT_TIMEOUT;

	for ( int i = 0; i < 16; i++ ) {								// read row into array

		if ( LCDbusyWait() == LCD_RESULT_TIMEOUT )
			return LCD_RESULT_TIMEOUT;

		LCD_PORT_DB :> void;										// prepare read operation
		LCD_PORT_RS <: LCD_TYPE_DATA;
		LCD_PORT_RW <: LCD_TYPE_READ;

		LCD_PORT_E <: 1;											// enable: pulse (rising edge)
		tmr :> t;													// wait: enable pulse width
		tmr when timerafter(t + LCD_TIMING_Tpw) :> void;

		LCD_PORT_DB :> row[i];										// save data

		LCD_PORT_E  <: 0;											// enable: clock (falling edge)

		tmr :> t;													// wait: remaining enable cycle time or timeout
			tmr when timerafter(t + LCD_TIMING_Tc - LCD_TIMING_Tpw) :> void;
	}

	if ( LCDcommand(0x01) == LCD_RESULT_TIMEOUT )					// clear display
		return LCD_RESULT_TIMEOUT;

	for ( int i = 0; i < 16; i++ ) {								// write array to top line without character translations
		if ( LCDbusyWait() == LCD_RESULT_TIMEOUT )
			return LCD_RESULT_TIMEOUT;
		else
			LCDoutRaw( LCD_TYPE_DATA, row[i] );
	}

	return LCD_RESULT_SUCCESS;
}


char LCDinit( void ) {
	// (re)initialise LCD display
	// Returns: LCD_RESULT_SUCCESS or LCD_RESULT_TIMEOUT
	timer tmr;
	int t;
	const unsigned wakeup_delay[] = { 500000, 16000, 16000 };		// time to wait after each 'wake up' command

	LCD_PORT_E  <: 0;

	tmr :> t;														// wait >15ms after power-up
	t += 1500000;
	tmr when timerafter(t) :> void;

	for( int i=0; i<3; i++ ) {										// send three 'wake up' instructions
		LCDoutRaw(LCD_TYPE_INSTR, 0x30);							// busy flag not available
		tmr :> t;													// wait: programmed delay
		t += wakeup_delay[i];
		tmr when timerafter(t) :> void;
	}

	if ( LCDcommand(0x38) == LCD_RESULT_TIMEOUT )					// function set: 8-bit/2-line
			return LCD_RESULT_TIMEOUT;
	if ( LCDcommand(0x10) == LCD_RESULT_TIMEOUT )					// set cursor
			return LCD_RESULT_TIMEOUT;
	if ( LCDcommand(0x0c) == LCD_RESULT_TIMEOUT )					// display ON, cursor OFF, blink OFF
			return LCD_RESULT_TIMEOUT;
	if ( LCDcommand(0x06) == LCD_RESULT_TIMEOUT )					// entry mode set: increment AH, display shift OFF
			return LCD_RESULT_TIMEOUT;

	return LCD_RESULT_SUCCESS;
}
